<?xml version="1.0" encoding="US-ASCII"?>
<!--
     Render a genericode file to a simple HTML presentation.

     This provides the basic mechanics so that users can embellish on this
     presentation for their own purposes.

     To enable a genericode file to be rendered in a browser,
     add the following processing instruction to the start of the file:

     <?xml-stylesheet type="text/xsl" href="Crane-genericode2html.xsl"?>

     When in a Windows interface, drag the genericode file from the Windows
     Explorer to Mozilla Firefox or Internet Explorer to engage the stylesheet
     presentation.

     For operation in Internet Explorer it is necessary in Windows to indicate
     that ".gc" files are XML documents.  This is done in Windows Explorer
     under the Tools/FolderOptions... menu, under the File Types tab, clicking
     the "New" button, setting the File Extension to "gc", then clicking the
     "Advanced" button to set the Asociated File Type to "XML Document".  Using
     the "Change" button next to the "Opens with:" herald, one can then set
     the recommended program to be Internet Explorer.

     When creating standalone HTML from the XML, two examples of using Saxon
     are as follows, the first explicitly referencing the stylesheet and the
     second looking for the embedded stylesheet association:

       java -jar saxon.jar -o file.html file.gc Crane-genericode2html.xsl
       java -jar saxon.jar -o file.html -a file.gc

     Note that it is not necessary for any referenced value list files to have
     an embedded stylesheet when using a stylesheet with a CIQ context/value
     association file.

 $Id: Crane-genericode2html.xsl,v 1.4 2006/12/09 19:00:09 G. Ken Holman Exp $

 Copyright (C) - Crane Softwrights Ltd.
               - http://www.CraneSoftwrights.com/links/res-dev.htm
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
 
 - Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
 - Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 - The name of the author may not be used to endorse or promote products
   derived from this software without specific prior written permission.
 
 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
 NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 Note: for your reference, the above is the "Modified BSD license", this text
     was obtained 2003-07-26 at http://www.xfree86.org/3.3.6/COPYRIGHT2.html#5
-->

<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:gc="http://docs.oasis-open.org/codelist/ns/genericode/1.0/"
                xmlns="http://www.w3.org/TR/REC-html40"
                version="1.0">

<!--========================================================================-->
<!--wrap processing of document element so that document element can be
    processed on its own in another context-->

<xsl:template match="/">
  <html>
    <head>
      <xsl:choose>
        <xsl:when test="gc:CodeList">
          <!--found what we are expecting-->
          <title>
            Enumeration: 
            ShortName=<xsl:value-of select="*/Identification/ShortName"/>
          </title>
        </xsl:when>
        <xsl:otherwise>
          <title>Unexpected XML document for stylesheet</title>
        </xsl:otherwise>
      </xsl:choose>
    </head>
    <body>
      <xsl:apply-templates/>
    </body>
  </html>
</xsl:template>

<xsl:template match="/*">
  <p>
 Unexpected XML document for <samp>Crane-genericode2html.xsl</samp> stylesheet.
  </p>
  <p>
    The document element is expected to be named 
    <samp>CodeList</samp> in the
    <samp><xsl:value-of select="document('')/*/namespace::gc"/></samp>
    namespace, otherwise this message is displayed.
  </p>
  <p>
    What was found in this document is the document element named
    <samp><xsl:value-of select="name(.)"/></samp> in the
    <samp><xsl:value-of select="namespace-uri(.)"/></samp> namespace.
  </p>
</xsl:template>

<!--========================================================================-->
<!--the titling of the value list-->

<xsl:template match="/gc:CodeList" priority="2">
  <!--when used from the UBL context/value association file stylesheet, the
      $uri parameter reflects the URI at which this file is being referenced-->
  <xsl:param name="uri"/>
  <h2>
    <a name="{generate-id(.)}">
      List
      <xsl:if test="$uri">
        <samp>uri=<xsl:value-of select="$uri"/></samp>
      </xsl:if>
    </a>
  </h2>
  <blockquote>
    <xsl:apply-templates select="Identification"/>
  </blockquote>
  <table border="1">
    <xsl:apply-templates select="ColumnSet"/>
    <xsl:apply-templates select="SimpleCodeList"/>
  </table>
</xsl:template>

<!--present the identification infromation in a table-->
<xsl:template match="Identification">
  <table>
    <xsl:apply-templates mode="gc:identification"/>
  </table>
</xsl:template>

<!--leaf-level identification information is in a cell-->
<xsl:template match="*" mode="gc:identification">
  <tr>
    <td valign="top">
      <samp>
        <xsl:value-of select="name(.)"/>
        <xsl:if test="@*">
          <xsl:text> (</xsl:text>
          <xsl:for-each select="@*">
            <xsl:if test="position()&gt;1"><xsl:text> </xsl:text></xsl:if>
            <xsl:value-of select="concat(name(.),'=&#34;',.,'&#34;')"/>
          </xsl:for-each>
          <xsl:text>)</xsl:text>
        </xsl:if>
      </samp>
    </td>
    <td valign="top"><samp>=</samp></td>
    <td valign="top"><samp><xsl:value-of select="."/></samp></td>
  </tr>
</xsl:template>

<!--nested identification information is in a nested table-->
<xsl:template match="*[*]" mode="gc:identification">
  <tr>
    <td valign="top">
      <samp>
        <xsl:value-of select="name(.)"/>
      </samp>
    </td>
    <td valign="top"><br/></td>
    <td valign="top">
      <table>
        <xsl:apply-templates mode="gc:identification"/>
      </table>
    </td>
  </tr>
</xsl:template>

<!--========================================================================-->
<!--the headings of the columns of values-->

<xsl:template match="ColumnSet">
  <tr>
    <xsl:apply-templates/>
  </tr>
</xsl:template>

<xsl:template match="Column">
  <th align="center" valign="top">
    <!--the column title-->
    <xsl:apply-templates select="ShortName"/>
    <br/>
    <!--show the meta data-->
    <samp>
      <small>
        <i>Id="<xsl:value-of select="@Id"/>"</i>
        <br/>
        <xsl:text/>(<xsl:value-of select="@Use"/>)<xsl:text/>
        <br/>
        <xsl:for-each select="Data/@*">
          <xsl:if test="position()&gt;1"><xsl:text> </xsl:text></xsl:if>
          <xsl:value-of select="concat(name(.),'=&#34;',.,'&#34;')"/>
        </xsl:for-each>
        <!--if this is a key column, indicate the key information-->
        <xsl:for-each select="../Key[ColumnRef/@Ref=current()/@Id]">
          <br/>
          <xsl:text>[Key Id="</xsl:text>
          <xsl:value-of select="concat(@Id,'&#34;: ',ShortName,']')"/>
        </xsl:for-each>
      </small>
    </samp>
  </th>
</xsl:template>

<xsl:template match="ColumnSet/Key"/>

<!--========================================================================-->
<!--the rows of values-->

<xsl:template match="Row">
  <xsl:variable name="thisRow" select="."/>
  <tr>
    <!--present the columns of the data in the order of the column set
        (which may be different)-->
    <xsl:for-each select="../../ColumnSet/Column/@Id">
      <xsl:apply-templates select="$thisRow/Value[@ColumnRef=current()]"/>
    </xsl:for-each>
    <!--no error checking that there are values in the row that do not
        correspond to the column sets-->
  </tr>
</xsl:template>

<xsl:template match="Value">
  <td align="center" valign="top">
    <xsl:apply-templates/>
  </td>
</xsl:template>

</xsl:stylesheet>
